#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>

#include "lib.h"
#include "common.h"
#include "libtar.h"

int mkpak(struct stream *out, char *name, char *version, char *description,
		struct dep *deps, struct stringlist *untracked, char *username,
		char *groupname, struct path_info **pi) {
	unsigned char block[512];
	struct path_info *cpi;
	struct dep *cd;
	struct passwd *pw;
	struct group *gr;
	uid_t uid;
	gid_t gid;
	size_t i, dsize;
	int result;
	if (username)
		uid = 65534;
	else
		uid = 0;
	if (groupname)
		gid = 65534;
	else
		gid = 0;
	for (i = 0; i < 512; i++)
		block[i] = 0;
	if (strlen(name) > 100) {
		fprintf(stderr, "%s: package name too long (max 100 characters)\n",
			progname);
		return -1;
	}
	if (strchr(name, '/')) {
		fprintf(stderr, "%s: package name contains /\n", progname);
		return -1;
	}
	if (strchr(name, 10)) {
		fprintf(stderr, "%s: package name contains newline\n", progname);
		return -1;
	}
	strncpy(block, name, 100);
	if (strlen(version) > 24) {
		fprintf(stderr, "%s: version string too long (max 24 characters)\n",
			progname);
		return -1;
	}
	if (strchr(name, 10)) {
		fprintf(stderr, "%s: version string contains newline\n", progname);
		return -1;
	}
	strncpy(block + 100, version, 24);
	dsize = 0;
	for (cd = deps; cd; cd = cd->next)
		dsize += strlen(cd->name) + strlen(cd->min_version) +
			strlen(cd->max_version) + 3;
	snprintf(block + 124, 12, "%011o", dsize);
	sprintf(block + 136, "00000000000");
	block[156] = '#';
	strncpy(block + 157, description, 355);
	snprintf(block + 148, 8, "%06o ", checksum(block, 0));
	x_stream_write(out, block, 512);
	for (cd = deps; cd; cd = cd->next) {
		x_stream_write(out, cd->name, strlen(cd->name) + 1);
		x_stream_write(out, cd->min_version, strlen(cd->min_version) + 1);
		x_stream_write(out, cd->max_version, strlen(cd->max_version) + 1);
	}
	fill_bytes(out, (512 - dsize % 512) % 512);
	if (username) {
		pw = getpwnam(username);
		if (pw)
			uid = pw->pw_uid;
		else
			fprintf(stderr, "%s: warning: invalid %s\n", progname, username);
	}
	if (groupname) {
		gr = getgrnam(groupname);
		if (gr)
			gid = gr->gr_gid;
		else
			fprintf(stderr, "%s: warning: invalid %s\n", progname, groupname);
	}
	result = 0;
	if (prepare_path(pi, NULL, ".", 0))
		result = -1;
	for (cpi = *pi; cpi; cpi = cpi->next)
		if (output_path(out, cpi, *pi, untracked, uid, username, gid,
				groupname))
			result = -1;
	output_footer(out);
	return result;
}

int main(int argc, char **argv) {
	struct dep *deps, *new, *cd;
	struct stream *file;
	struct path_info *pi;
	struct stringlist *untracked, *nsl;
	char *filename, *name, *ver, *description, *username, *groupname;
	int i, r;
	progname = argv[0];
	filename = NULL;
	name = NULL;
	ver = NULL;
	description = NULL;
	deps = NULL;
	cd = NULL;
	new = NULL;
	username = NULL;
	groupname = NULL;
	pi = NULL;
	untracked = NULL;
	for (i = 1; i < argc; i++) {
		if (argv[i][0] == '-' && argv[i][1]) {
			if (argv[i][2]) {
				fprintf(stderr, "%s: invalid %s (%s -? for usage)\n",
					progname, argv[i], progname);
				exit(1);
			}
			switch (argv[i][1]) {
			case 'n':
				if (i + 1 >= argc || argv[i+1][0] == '-') {
					fprintf(stderr,
						"%s: -n requires argument (%s -? for usage)\n",
						progname, progname);
					exit(1);
				}
				i++;
				name = argv[i];
				break;
			case 'v':
				if (i + 1 >= argc || argv[i+1][0] == '-') {
					fprintf(stderr,
						"%s: -v requires argument (%s -? for usage)\n",
						progname, progname);
					exit(1);
				}
				i++;
				ver = argv[i];
				break;
			case 'd':
				if (i + 1 >= argc || argv[i+1][0] == '-') {
					fprintf(stderr,
						"%s: -d requires argument (%s -? for usage)\n",
						progname, progname);
					exit(1);
				}
				i++;
				description = argv[i];
				break;
			case 'D':
				if (i + 1 >= argc || argv[i+1][0] == '-') {
					fprintf(stderr,
						"%s: -D requires argument (%s -? for usage)\n",
						progname, progname);
					exit(1);
				}
				i++;
				new = malloc(sizeof(struct dep));
				if (!new) {
					fprintf(stderr, "%s: out of memory\n");
					exit(1);
				}
				new->name = argv[i];
				new->min_version = "";
				new->max_version = "";
				new->next = NULL;
				if (!cd)
					deps = new;
				else
					cd->next = new;
				cd = new;
				break;
			case 'm':
				if (i + 1 >= argc || argv[i+1][0] == '-') {
					fprintf(stderr,
						"%s: -m requires argument (%s -? for usage)\n",
						progname, progname);
					exit(1);
				}
				i++;
				if (!deps) {
					fprintf(stderr,
						"%s: -m must follow -D (%s -? for usage)\n", progname,
						progname);
					exit(1);
				}
				deps->min_version = argv[i];
				break;
			case 'M':
				if (i + 1 >= argc || argv[i+1][0] == '-') {
					fprintf(stderr,
						"%s: -M requires argument (%s -? for usage)\n",
						progname, progname);
					exit(1);
				}
				i++;
				if (!deps) {
					fprintf(stderr,
						"%s: -M must follow -D (%s -? for usage\n", progname,
						progname);
					exit(1);
				}
				deps->max_version = argv[i];
				break;
			case 'o':
				if (i + 1 >= argc || argv[i+1][0] == '-') {
					fprintf(stderr,
						"%s: -o requires argument (%s -? for usage)\n",
						progname, progname);
					exit(1);
				}
				i++;
				username = argv[i];
				break;
			case 'g':
				if (i + 1 >= argc || argv[i+1][0] == '-') {
					fprintf(stderr,
						"%s: -g requires argument (%s -? for usage)\n",
						progname, progname);
					exit(1);
				}
				i++;
				groupname = argv[i];
				break;
			case 'u':
				if (i + 1 >= argc || argv[i+1][0] == '-') {
					fprintf(stderr,
						"%s: -u requires argument (%s -? for usage)\n",
						progname, progname);
					exit(1);
				}
				i++;
				nsl = malloc(sizeof(struct stringlist));
				if (!nsl) {
					fprintf(stderr, "%s: out of memory\n", progname);
					exit(1);
				}
				nsl->next = untracked;
				nsl->s = argv[i];
				untracked = nsl;
				break;
			case 'V':
				version("mkpak");
				return 0;
			case '?':
				version("mkpak");
				fprintf(stderr,
					"\n%s OPTIONS FILENAME\n"
					"\tCreate a new package file named FILENAME described by\n"
					"\tOPTIONS and containing the contents of the current\n"
					"\tdirectory as if it were the root directory.\n"
					"\t-n NAME\tRequired. Make the name of the package NAME.\n"
					"\t-v VERSION\tRequired. Make the version of the package\n"
					"\t\t\tVERSION.\n"
					"\t-d DESCRIPTION\tMake the description of the package\n"
					"\t\t\tDESCRIPTION.\n"
					"\t-D DEPENDENCY\tRequire a package named DEPENDENCY as\n"
					"\t\t\ta dependency of the package.\n"
					"\t-m VERSION\tRequire at a minimum version VERSION of\n"
					"\t\t\tthe most recently listed dependency.\n"
					"\t-M VERSION\tRequire at a maximum version VERSION of\n"
					"\t\t\tthe most recently listed dependency.\n"
					"\t-o OWNER\tSet the owner of all files in the package\n"
					"\t\t\tto OWNER.\n"
					"\t-g GROUP\tSet the group of all files in the package\n"
					"\t\t\tto GROUP.\n"
					"\t-u PATH\tIf PATH or a descendent is included in the\n"
					"\t\t\tpackage, do not uninstall it when the package is\n"
					"\t\t\tuninstalled.\n"
					"%s -V\n"
					"\tDisplay the version of this command.\n"
					"%s -?\n"
					"\tDisplay this usage message.\n",
					progname, progname, progname);
				return 0;
			default:
				fprintf(stderr, "%s: invalid -%c (%s -? for usage)\n",
					progname, progname);
				exit(1);
			}
		} else {
			if (filename) {
				fprintf(stderr,
					"%s: invalid %s: too many filenames (%s -? for usage)\n",
					progname, argv[i], progname);
				exit(1);
			}
			filename = argv[i];
		}
	}
	if (!filename) {
		fprintf(stderr, "%s: filename required (%s -? for usage)\n", progname,
			progname);
		exit(1);
	}
	if (!name) {
		fprintf(stderr, "%s: -n required (%s -? for usage)\n", progname,
			progname);
		exit(1);
	}
	if (!ver) {
		fprintf(stderr, "%s: -v required (%s -? for usage)\n", progname,
			progname);
		exit(1);
	}
	if (!description)
		description = "";
	file = xz_stream_open(filename, "w");
	r = mkpak(file, name, ver, description, deps, untracked, username,
		groupname, &pi);
	file->close(file);
	return -r;
}
